/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_SLICE_H_
#define AIDGE_CORE_OPERATOR_SLICE_H_

#include <memory>
#include <vector>

#include "aidge/backend/OperatorImpl.hpp"
#include "aidge/data/Tensor.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/operator/OperatorTensor.hpp"
#include "aidge/utils/Registrar.hpp"
#include "aidge/utils/StaticAttributes.hpp"
#include "aidge/utils/Types.h"

namespace Aidge {
enum class SliceAttr { Beginning, SliceDims };

class Slice_Op
    : public OperatorTensor,
      public Registrable<Slice_Op, std::string, std::unique_ptr<OperatorImpl>(const Slice_Op &)>,
      public StaticAttributes<SliceAttr, std::size_t, std::vector<DimSize_t>> {
public:
    static const std::string Type;

    Slice_Op() = delete;

    using Attributes_ = StaticAttributes<SliceAttr, std::size_t, std::vector<DimSize_t>>;
    template <SliceAttr e>
    using attr = typename Attributes_::template attr<e>;

    Slice_Op(const std::size_t beginningPos, const std::vector<DimSize_t> sliceDims)
        : OperatorTensor(Type, 1, 0, 1),
          Attributes_(attr<SliceAttr::Beginning>(beginningPos),
                      attr<SliceAttr::SliceDims>(sliceDims))
    {}

    /**
     * @brief Copy-constructor. Copy the operator attributes and its output tensor(s), but not its
     * input tensors (the new operator has no input associated).
     * @param op Operator to copy.
     */
    Slice_Op(const Slice_Op &op)
        : OperatorTensor(op),
          Attributes_(op)
    {
        mImpl = op.mImpl ? Registrar<Slice_Op>::create(op.mOutputs[0]->getImpl()->backend())(*this)
                         : nullptr;
    }

public:
    /**
     * @brief Clone the operator using its copy-constructor.
     * @see Operator::Slice_Op
     */
    std::shared_ptr<Operator> clone() const override { return std::make_shared<Slice_Op>(*this); }

    void computeOutputDims() override final {
        if (!getInput(0) || (getInput(0)->empty())) {
            AIDGE_THROW_OR_ABORT(std::runtime_error, "Every input should be associated with a Tensor");
        }
        std::vector<DimSize_t> outputDims = std::vector<DimSize_t>(getInput(0)->nbDims());
        const std::vector<DimSize_t> inputDims = getInput(0)->dims();

        // Check that the sliced Tensor is actually part of the input Tensor
        // For a 5*5 tensor ('x') and a 3*3 slice kernel ('o'):
        // xxxxx               xxxxx
        // xxxxx               xxxxx
        // xxooo  --> ok       xxxoo --> out of bound
        // xxooo               xxxoo
        // xxooo               xxxoo
        std::vector<std::size_t> beginningCoords = mInputs[0]->getCoord(this->template getAttr<SliceAttr::Beginning>());
        for (std::size_t i = 0; i < getInput(0)->nbDims(); ++i) {
            if (beginningCoords[i] + this->template getAttr<SliceAttr::SliceDims>()[i] > inputDims[i]) {
                AIDGE_THROW_OR_ABORT(std::runtime_error, "ROI of Slice operator out of bounds");
            } else {
                outputDims[i] = this->template getAttr<SliceAttr::SliceDims>()[i];
            }
        }
        mOutputs[0]->resize(outputDims);
    }

    void setBackend(const std::string &name, int device = 0) override {
        mImpl = Registrar<Slice_Op>::create(name)(*this);
        mOutputs[0]->setBackend(name, device);
    }

    static const std::vector<std::string> getInputsName(){
        return {"data_input"};
    }
    static const std::vector<std::string> getOutputsName(){
        return {"data_output"};
    }
};


inline std::shared_ptr<Node> Slice(const std::size_t beginningPos, const std::vector<DimSize_t> sliceDims,
                                   const std::string &name = "") {
    // FIXME: properly handle default w&b initialization in every cases
    return std::make_shared<Node>(std::make_shared<Slice_Op>(beginningPos, sliceDims), name);
}
}  // namespace Aidge

namespace {
template <>
const char *const EnumStrings<Aidge::SliceAttr>::data[] = { "Beginning", "SliceDims" };
}

#endif /* AIDGE_CORE_OPERATOR_RELU_H_ */
